@@ -12,7 +12,8 @@ except ImportError:
 from makefun import create_function, add_signature_parameters, remove_signature_parameters, with_signature, wraps, \
     create_wrapper
 
-python_version = sys.version_info.major
+
+PY2 = sys.version_info < (3,)
 
 
 @pytest.mark.parametrize('decorator', [False, True], ids="decorator={}".format)
@@ -328,3 +329,38 @@ def test_wraps_remove():
         return a + b
 
     assert 12 <= summer(b=12) <= 13
+
+
+@pytest.mark.parametrize("prepend", [True, False], ids="prepend={}".format)
+def test_wraps_add(prepend):
+
+    def add_a_to_result(f):
+        """
+        A decorator that injects a random number inside the `a` argument,
+        removing it from the exposed signature
+        """
+        if prepend:
+            decorator = wraps(f, prepend_args='a')
+        else:
+            decorator = wraps(f, append_args='a')
+
+        if PY2:
+            def my_wrapper(*args, **kwargs):
+                a = kwargs.pop('a')
+                return a + f(*args, **kwargs)
+        else:
+            from ._test_py35 import get_my_wrapper
+            my_wrapper = get_my_wrapper(f)
+
+        my_wrapper = decorator(my_wrapper)
+
+        return my_wrapper
+
+    @add_a_to_result
+    def identity(b):
+        return b
+
+    assert identity(b=12, a=0.5) == 12.5
+
+    ref_str = "(a, b)" if prepend else "(b, a)"
+    assert str(signature(identity)) == ref_str
